# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

import uuid
from unittest.mock import MagicMock, patch

import pytz
from flask import Flask
from invenio_db import db
from invenio_records import Record
from invenio_search.engine import check_os_version

from invenio_indexer import InvenioIndexer
from invenio_indexer.api import RecordIndexer
from invenio_indexer.signals import before_record_index

_global_magic_hook = MagicMock()
"""Iternal importable magic hook instance."""


def test_version():
    """Test version import."""
    from invenio_indexer import __version__

    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask("testapp")
    ext = InvenioIndexer(app)
    assert "invenio-indexer" in app.extensions

    app = Flask("testapp")
    ext = InvenioIndexer()
    assert "invenio-indexer" not in app.extensions
    ext.init_app(app)
    assert "invenio-indexer" in app.extensions


def test_hook_initialization(base_app):
    """Test hook initialization."""
    app = base_app
    magic_hook = MagicMock()
    app.config["INDEXER_BEFORE_INDEX_HOOKS"] = [
        magic_hook,
        "test_invenio_indexer:_global_magic_hook",
    ]
    ext = InvenioIndexer(app)

    with app.app_context():
        recid = uuid.uuid4()
        record = Record.create({"title": "Test"}, id_=recid)
        db.session.commit()

        client_mock = MagicMock()
        RecordIndexer(search_client=client_mock, version_type="force").index(record)
        args = (app,)
        kwargs = dict(
            index=app.config["INDEXER_DEFAULT_INDEX"],
            arguments={},
            record=record,
            json={
                "title": "Test",
                "_created": pytz.utc.localize(record.created).isoformat(),
                "_updated": pytz.utc.localize(record.updated).isoformat(),
            },
        )
        magic_hook.assert_called_with(*args, **kwargs)
        _global_magic_hook.assert_called_with(*args, **kwargs)
        client_mock.index.assert_called_with(
            id=str(recid),
            version=0,
            version_type="force",
            index=app.config["INDEXER_DEFAULT_INDEX"],
            body={
                "title": "Test",
                "_created": pytz.utc.localize(record.created).isoformat(),
                "_updated": pytz.utc.localize(record.updated).isoformat(),
            },
        )


def test_index_prefixing(base_app):
    """Test index prefixing."""
    app = base_app
    app.config["INDEXER_REPLACE_REFS"] = False
    app.config["SEARCH_INDEX_PREFIX"] = "test-"
    ext = InvenioIndexer(app)

    default_index = app.config["INDEXER_DEFAULT_INDEX"]

    with app.app_context():
        with patch("invenio_records.api._records_state.validate"):
            record = Record.create({"title": "Test"})
            record2 = Record.create(
                {
                    "$schema": "/records/authorities/authority-v1.0.0.json",
                    "title": "Test with schema",
                }
            )
            record3 = Record.create({"title": "Test"})
            record3.delete()
        db.session.commit()

        with patch(
            "invenio_indexer.signals.before_record_index.send"
        ) as before_record_index_send:
            client_mock = MagicMock()
            RecordIndexer(search_client=client_mock).index(record)
            client_mock.index.assert_called_with(
                id=str(record.id),
                version=0,
                version_type="external_gte",
                index="test-" + default_index,
                body={
                    "title": "Test",
                    "_created": pytz.utc.localize(record.created).isoformat(),
                    "_updated": pytz.utc.localize(record.updated).isoformat(),
                },
            )
            before_record_index_send.assert_called_with(
                app,
                json={
                    "title": "Test",
                    "_created": pytz.utc.localize(record.created).isoformat(),
                    "_updated": pytz.utc.localize(record.updated).isoformat(),
                },
                record=record,
                index=default_index,  # non-prefixed index passed to receiver
                arguments={},
            )

            RecordIndexer(search_client=client_mock).index(record2)
            client_mock.index.assert_called_with(
                id=str(record2.id),
                version=0,
                version_type="external_gte",
                index="test-records-authorities-authority-v1.0.0",
                body={
                    "$schema": "/records/authorities/authority-v1.0.0.json",
                    "title": "Test with schema",
                    "_created": pytz.utc.localize(record2.created).isoformat(),
                    "_updated": pytz.utc.localize(record2.updated).isoformat(),
                },
            )
            before_record_index_send.assert_called_with(
                app,
                json={
                    "$schema": "/records/authorities/authority-v1.0.0.json",
                    "title": "Test with schema",
                    "_created": pytz.utc.localize(record2.created).isoformat(),
                    "_updated": pytz.utc.localize(record2.updated).isoformat(),
                },
                record=record2,
                index="records-authorities-authority-v1.0.0",  # no prefix
                arguments={},
            )
            RecordIndexer(search_client=client_mock).delete(record3)
            client_mock.delete.assert_called_with(
                id=str(record3.id),
                index="test-" + default_index,
                version=record3.revision_id,
                version_type="external_gte",
            )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration."""

import os
import shutil
import tempfile

import pytest
from celery.messaging import establish_connection
from flask import Flask
from flask_celeryext import FlaskCeleryExt
from invenio_db import InvenioDB, db
from invenio_records import InvenioRecords
from invenio_search import InvenioSearch
from invenio_search.engine import dsl
from sqlalchemy_utils.functions import create_database, database_exists, drop_database

from invenio_indexer import InvenioIndexer


@pytest.fixture()
def base_app(request):
    """Base application fixture."""
    instance_path = tempfile.mkdtemp()
    app = Flask("testapp", instance_path=instance_path)
    app.config.update(
        CELERY_BROKER_URL=os.environ.get(
            "BROKER_URL", "amqp://guest:guest@localhost:5672//"
        ),
        CELERY_TASK_ALWAYS_EAGER=True,
        CELERY_CACHE_BACKEND="memory",
        CELERY_TASK_EAGER_PROPAGATES=True,
        CELERY_RESULT_BACKEND="cache",
        INDEXER_DEFAULT_INDEX="records-default-v1.0.0",
        SQLALCHEMY_DATABASE_URI=os.environ.get(
            "SQLALCHEMY_DATABASE_URI", "sqlite:///test.db"
        ),
        SQLALCHEMY_TRACK_MODIFICATIONS=False,
        TESTING=True,
    )
    FlaskCeleryExt(app)
    InvenioDB(app)
    InvenioRecords(app)
    search = InvenioSearch(app, entry_point_group=None)
    search.register_mappings("records", "data")

    with app.app_context():
        if not database_exists(str(db.engine.url)):
            create_database(str(db.engine.url))
        db.create_all()

    def teardown():
        with app.app_context():
            drop_database(str(db.engine.url))
        shutil.rmtree(instance_path)

    request.addfinalizer(teardown)
    return app


@pytest.fixture()
def app(base_app):
    """Flask application fixture."""
    InvenioIndexer(base_app)
    return base_app


@pytest.fixture()
def queue(app):
    """Get queue object for testing bulk operations."""
    queue = app.config["INDEXER_MQ_QUEUE"]

    with app.app_context():
        with establish_connection() as c:
            q = queue(c)
            q.declare()
            q.purge()

    return queue


@pytest.fixture(scope="function")
def search_prefix(app):
    """Sets the search prefix."""
    tmp = app.config["SEARCH_INDEX_PREFIX"]
    app.config["SEARCH_INDEX_PREFIX"] = "test-"
    yield
    app.config["SEARCH_INDEX_PREFIX"] = tmp


@pytest.fixture(scope="function")
def record_cls_with_index(app):
    """Record class with index field."""

    class Record:
        """Simulates a record class with an IndexField.

        Since the IndexField has the __get__ method overwritten
        to return the actual index object this class behaves in
        the same way.
        """

        index = dsl.Index(app.config["INDEXER_DEFAULT_INDEX"])

    return Record



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test API."""

from unittest.mock import patch

import pytest

from invenio_indexer.utils import schema_to_index


def test_schema_to_index_with_names(app):
    """Test that prefix is added to the index when creating it."""
    result = schema_to_index("default.json", index_names=["default"])
    assert result == "default"


@pytest.mark.parametrize(
    ("schema, expected, index_names"),
    (
        (
            "records/record-v1.0.0.json",
            "records-record-v1.0.0",
            None,
        ),
        (
            "/records/record-v1.0.0.json",
            "records-record-v1.0.0",
            None,
        ),
        (
            "default-v1.0.0.json",
            "default-v1.0.0",
            None,
        ),
        (
            "default-v1.0.0.json",
            None,
            [],
        ),
        (
            "invalidextension",
            None,
            None,
        ),
    ),
)
def test_schema_to_index(schema, expected, index_names, app):
    """Test the expected value of schema to index."""
    result = schema_to_index(schema, index_names=index_names)
    assert result == expected



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2024 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Utility functions for data processing."""

import os
from functools import wraps

from flask import current_app
from invenio_search import current_search
from invenio_search.utils import build_index_from_parts


def schema_to_index(schema, index_names=None):
    """Get index given a schema URL.

    :param schema: The schema name
    :param index_names: A list of index name.
    :returns: The index.
    """
    parts = schema.split("/")
    rec_type, ext = os.path.splitext(parts[-1])
    parts[-1] = rec_type

    if ext not in {
        ".json",
    }:
        return None

    if index_names is None:
        index = build_index_from_parts(*parts)
        return index

    for start in range(len(parts)):
        name = build_index_from_parts(*parts[start:])
        if name in index_names:
            return name

    return None


def default_record_to_index(record):
    """Get index given a record.

    It tries to extract from `record['$schema']` the index.
    If it fails, return the default values.

    :param record: The record object.
    :returns: index.
    """
    # The indices could be defined either in the mappings or in the index_templates
    index_names = list(current_search.mappings.keys()) + list(
        current_search.index_templates.keys()
    )
    schema = record.get("$schema", "")
    if isinstance(schema, dict):
        schema = schema.get("$ref", "")

    index = schema_to_index(schema, index_names=index_names)

    return index or current_app.config["INDEXER_DEFAULT_INDEX"]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test API."""

import json
import uuid
from unittest.mock import MagicMock, patch

import pytz
from celery.messaging import establish_connection
from invenio_db import db
from invenio_records.api import Record
from invenio_search.engine import dsl
from jsonresolver import JSONResolver
from jsonresolver.contrib.jsonref import json_loader_factory
from kombu.compat import Consumer

from invenio_indexer.api import BulkRecordIndexer, RecordIndexer
from invenio_indexer.signals import before_record_index


def test_indexer_bulk_index(app, queue):
    """Test delay indexing."""
    with app.app_context():
        with establish_connection() as c:
            indexer = RecordIndexer()
            id1 = uuid.uuid4()
            id2 = uuid.uuid4()
            indexer.bulk_index([id1, id2])
            indexer.bulk_delete([id1, id2])

            consumer = Consumer(
                connection=c,
                queue=indexer.mq_queue.name,
                exchange=indexer.mq_exchange.name,
                routing_key=indexer.mq_routing_key,
            )

            messages = list(consumer.iterqueue())
            [m.ack() for m in messages]

            assert len(messages) == 4
            data0 = messages[0].decode()
            assert data0["id"] == str(id1)
            assert data0["op"] == "index"
            data2 = messages[2].decode()
            assert data2["id"] == str(id1)
            assert data2["op"] == "delete"


def test_delete_action(app):
    """Test delete action."""
    with app.app_context():
        testid = str(uuid.uuid4())
        action = RecordIndexer()._delete_action(
            dict(id=testid, op="delete", index="idx")
        )
        assert action["_op_type"] == "delete"
        assert action["_index"] == "idx"
        assert action["_id"] == testid

        # Skip JSONSchema validation
        with patch("invenio_records.api._records_state.validate"):
            record = Record.create(
                {
                    "$schema": {"$ref": "/records/authorities/authority-v1.0.0.json"},
                    "title": "Test",
                }
            )
            db.session.commit()
        action = RecordIndexer()._delete_action(
            dict(id=str(record.id), op="delete", index=None)
        )
        assert action["_op_type"] == "delete"
        assert action["_index"] == "records-authorities-authority-v1.0.0"
        assert action["_id"] == str(record.id)

        record.delete()
        db.session.commit()
        action = RecordIndexer()._delete_action(
            dict(id=str(record.id), op="delete", index=None)
        )
        assert action["_op_type"] == "delete"
        # Deleted record doesn't have '$schema', so index cannot
        # be determined, resulting to the default from config
        assert action["_index"] == app.config["INDEXER_DEFAULT_INDEX"]
        assert action["_id"] == str(record.id)


def test_index_action(app):
    """Test index action."""
    with app.app_context():
        record = Record.create({"title": "Test"})
        db.session.commit()

        def receiver(sender, json=None, record=None, arguments=None, **kwargs):
            json["extra"] = "extra"
            arguments["pipeline"] = "foobar"

        with before_record_index.connected_to(receiver):
            action = RecordIndexer()._index_action(
                dict(
                    id=str(record.id),
                    op="index",
                )
            )
            assert action["_op_type"] == "index"
            assert action["_index"] == app.config["INDEXER_DEFAULT_INDEX"]
            assert action["_id"] == str(record.id)
            assert action["_version"] == record.revision_id
            assert action["_version_type"] == "external_gte"
            assert action["pipeline"] == "foobar"
            assert "title" in action["_source"]
            assert "extra" in action["_source"]


def test_process_bulk_queue(app, queue):
    """Test process indexing."""
    with app.app_context():
        # Create a test record
        r = Record.create({"title": "test"})
        db.session.commit()
        invalid_id2 = uuid.uuid4()

        RecordIndexer().bulk_index([r.id, invalid_id2])
        RecordIndexer().bulk_delete([r.id, invalid_id2])

        ret = {}

        def _mock_bulk(client, actions_iterator, **kwargs):
            ret["actions"] = list(actions_iterator)
            return len(ret["actions"])

        with patch("invenio_indexer.api.bulk", _mock_bulk):
            # Invalid actions are rejected
            assert RecordIndexer().process_bulk_queue() == 2
            assert [x["_op_type"] for x in ret["actions"]] == ["index", "delete"]


def test_process_bulk_queue_errors(app, queue):
    """Test error handling during indexing."""
    with app.app_context():
        # Create a test record
        r1 = Record.create({"title": "invalid", "reffail": {"$ref": "#/invalid"}})
        r2 = Record.create(
            {
                "title": "valid",
            }
        )
        db.session.commit()

        RecordIndexer().bulk_index([r1.id, r2.id])

        ret = {}

        def _mock_bulk(client, actions_iterator, **kwargs):
            ret["actions"] = list(actions_iterator)
            return len(ret["actions"])

        with patch("invenio_indexer.api.bulk", _mock_bulk):
            # Exceptions are caught
            assert RecordIndexer().process_bulk_queue() == 1
            assert len(ret["actions"]) == 1
            assert ret["actions"][0]["_id"] == str(r2.id)


def test_index(app):
    """Test record indexing."""
    with app.app_context():
        recid = uuid.uuid4()
        record = Record.create({"title": "Test"}, id_=recid)
        db.session.commit()

        client_mock = MagicMock()
        RecordIndexer(search_client=client_mock, version_type="force").index(
            record, arguments={"pipeline": "foobar"}
        )

        client_mock.index.assert_called_with(
            id=str(recid),
            version=0,
            version_type="force",
            index=app.config["INDEXER_DEFAULT_INDEX"],
            body={
                "title": "Test",
                "_created": pytz.utc.localize(record.created).isoformat(),
                "_updated": pytz.utc.localize(record.updated).isoformat(),
            },
            pipeline="foobar",
        )

        with patch("invenio_indexer.api.RecordIndexer.index") as fun:
            RecordIndexer(search_client=client_mock).index_by_id(recid)
            assert fun.called


def test_refresh_with_indexer(app, record_cls_with_index):
    """Test index refresh."""
    with app.app_context():
        client_mock = MagicMock()
        ri = RecordIndexer(
            search_client=client_mock,
            record_cls=record_cls_with_index,
            version_type="force",
        )
        ri.refresh()

        client_mock.indices.refresh.assert_called_with(
            index=app.config["INDEXER_DEFAULT_INDEX"]
        )


def test_refresh_with_indexer_and_prefix(search_prefix, app, record_cls_with_index):
    """Test index refresh."""
    with app.app_context():
        client_mock = MagicMock()
        ri = RecordIndexer(
            search_client=client_mock,
            record_cls=record_cls_with_index,
            version_type="force",
        )
        ri.refresh()

        prefix = app.config["SEARCH_INDEX_PREFIX"]
        index_name = app.config["INDEXER_DEFAULT_INDEX"]
        prefixed_index = f"{prefix}{index_name}"

        client_mock.indices.refresh.assert_called_with(index=prefixed_index)


def test_refresh_with_index_name(app):
    """Test index refresh."""
    with app.app_context():
        client_mock = MagicMock()
        index_name = app.config["INDEXER_DEFAULT_INDEX"]

        ri = RecordIndexer(search_client=client_mock, version_type="force")
        ri.refresh(index=index_name)

        client_mock.indices.refresh.assert_called_with(index=index_name)


def test_refresh_with_index_obj(app):
    """Test index refresh."""
    with app.app_context():
        client_mock = MagicMock()
        index_name = app.config["INDEXER_DEFAULT_INDEX"]

        ri = RecordIndexer(search_client=client_mock, version_type="force")
        ri.refresh(index=dsl.Index(index_name))

        client_mock.indices.refresh.assert_called_with(index=index_name)


def test_delete(app):
    """Test record indexing."""
    with app.app_context():
        recid = uuid.uuid4()
        record = Record.create({"title": "Test"}, id_=recid)
        db.session.commit()

        client_mock = MagicMock()
        RecordIndexer(search_client=client_mock).delete(record)

        client_mock.delete.assert_called_with(
            id=str(recid),
            index=app.config["INDEXER_DEFAULT_INDEX"],
            version=record.revision_id,
            version_type="external_gte",
        )

        with patch("invenio_indexer.api.RecordIndexer.delete") as fun:
            RecordIndexer(search_client=client_mock).delete_by_id(recid)
            assert fun.called


def test_replace_refs(app):
    """Test replace refs."""
    app.config["INDEXER_REPLACE_REFS"] = False
    app.extensions["invenio-records"].loader_cls = json_loader_factory(
        JSONResolver(plugins=["demo.json_resolver"])
    )

    with app.app_context():
        record = Record({"$ref": "http://dx.doi.org/10.1234/foo"})
        data = RecordIndexer()._prepare_record(record, "records", "record")
        assert "$ref" in data

    app.config["INDEXER_REPLACE_REFS"] = True
    with app.app_context():
        record = Record({"$ref": "http://dx.doi.org/10.1234/foo"})
        data = RecordIndexer()._prepare_record(record, "records", "record")
        assert "$ref" not in data
        assert json.dumps(data)


def test_bulkrecordindexer_index_delete_by_record_id(app, queue):
    """Test utility class BulkRecordIndexer index/delete by record id."""
    with app.app_context():
        with establish_connection() as c:
            indexer = BulkRecordIndexer()
            id1 = uuid.uuid4()
            indexer.index_by_id(id1)
            indexer.delete_by_id(id1)

            consumer = Consumer(
                connection=c,
                queue=indexer.mq_queue.name,
                exchange=indexer.mq_exchange.name,
                routing_key=indexer.mq_routing_key,
            )

            messages = list(consumer.iterqueue())
            [m.ack() for m in messages]

            assert len(messages) == 2
            data0 = messages[0].decode()
            assert data0["id"] == str(id1)
            assert data0["op"] == "index"
            data1 = messages[1].decode()
            assert data1["id"] == str(id1)
            assert data1["op"] == "delete"


def test_bulkrecordindexer_index_delete_by_record(app, queue):
    """Test utility class BulkRecordIndexer index/delete by record object."""
    with app.app_context():
        with establish_connection() as c:
            recid = uuid.uuid4()
            record = Record.create({"title": "Test"}, id_=recid)
            db.session.commit()
            indexer = BulkRecordIndexer()
            indexer.index(record)
            indexer.delete(record)

            consumer = Consumer(
                connection=c,
                queue=indexer.mq_queue.name,
                exchange=indexer.mq_exchange.name,
                routing_key=indexer.mq_routing_key,
            )

            messages = list(consumer.iterqueue())
            [m.ack() for m in messages]

            assert len(messages) == 2
            data0 = messages[0].decode()
            assert data0["id"] == str(recid)
            assert data0["op"] == "index"
            data1 = messages[1].decode()
            assert data1["id"] == str(recid)
            assert data1["op"] == "delete"


def test_before_record_index_dynamic_connect(app):
    """Test before_record_index.dynamic_connect."""
    with app.app_context():
        with patch("invenio_records.api._records_state.validate"):
            auth_record = Record.create(
                {
                    "$schema": "/records/authorities/authority-v1.0.0.json",
                    "title": "Test",
                }
            )
            bib_record = Record.create(
                {
                    "$schema": "/records/bibliographic/bibliographic-v1.0.0.json",
                    "title": "Test",
                }
            )
            db.session.commit()

        def _simple(sender, json=None, **kwargs):
            json["simple"] = "simple"

        def _custom(sender, json=None, **kwargs):
            json["custom"] = "custom"

        def _cond(sender, connect_kwargs, index=None, **kwargs):
            return "bibliographic" in index

        _receiver1 = before_record_index.dynamic_connect(
            _simple, index="records-authorities-authority-v1.0.0"
        )
        _receiver2 = before_record_index.dynamic_connect(_custom, condition_func=_cond)

        action = RecordIndexer()._index_action(dict(id=str(auth_record.id), op="index"))
        assert "title" in action["_source"]
        assert action["_source"]["simple"] == "simple"

        action = RecordIndexer()._index_action(
            dict(id=str(bib_record.id), index="foo", op="index")
        )
        assert "title" in action["_source"]
        assert action["_source"]["custom"] == "custom"

        before_record_index.disconnect(_receiver1)
        before_record_index.disconnect(_receiver2)


def test_indexer_record_class():
    """Tests the usage of a custom record class."""

    class DummyRecord:
        """Dummy record class."""

        pass

    indexer = RecordIndexer()
    assert indexer.record_cls == Record

    indexer = RecordIndexer(record_cls=DummyRecord)
    assert indexer.record_cls == DummyRecord



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""CLI for indexer."""

import click
from celery.messaging import establish_connection
from flask import current_app
from flask.cli import with_appcontext
from invenio_pidstore.models import PersistentIdentifier, PIDStatus
from invenio_search.cli import index

from .api import RecordIndexer
from .proxies import current_indexer_registry
from .tasks import process_bulk_queue


def abort_if_false(ctx, param, value):
    """Abort command is value is False."""
    if not value:
        ctx.abort()


def resultcallback(group):
    """Compatibility layer for Click 7 and 8."""
    if hasattr(group, "result_callback") and group.result_callback is not None:
        decorator = group.result_callback()
    else:
        # Click < 8.0
        decorator = group.resultcallback()
    return decorator


@index.command()
@click.option("--delayed", "-d", is_flag=True, help="Run indexing in background.")
@click.option(
    "--concurrency",
    "-c",
    default=1,
    type=int,
    help="Number of concurrent indexing tasks to start.",
)
@click.option(
    "--queue",
    "-q",
    type=str,
    help="Name of the celery queue used to put the tasks into.",
)
@click.option("--version-type", help="The search engine version type to use.")
@click.option(
    "--raise-on-error/--skip-errors",
    default=True,
    help="Controls if the search engine bulk indexing errors raise an exception.",
)
@click.option(
    "--chunk_size",
    default=500,
    type=int,
    help="Number of documents to send per batch.",
)
@with_appcontext
def run(
    delayed, concurrency, chunk_size, version_type=None, queue=None, raise_on_error=True
):
    """Run bulk record indexing."""
    if delayed:
        celery_kwargs = {
            "kwargs": {
                "version_type": version_type,
                "search_bulk_kwargs": {"raise_on_error": raise_on_error},
            }
        }
        click.secho(
            "Starting {0} tasks for indexing records...".format(concurrency), fg="green"
        )
        if queue is not None:
            celery_kwargs.update({"queue": queue})
        for c in range(0, concurrency):
            process_bulk_queue.apply_async(**celery_kwargs)
    else:
        click.secho("Indexing records...", fg="green")
        RecordIndexer(version_type=version_type).process_bulk_queue(
            search_bulk_kwargs={
                "raise_on_error": raise_on_error,
                "chunk_size": chunk_size,
            }
        )


@index.command()
@click.option(
    "--yes-i-know",
    is_flag=True,
    callback=abort_if_false,
    expose_value=False,
    prompt="Do you really want to reindex all records?",
)
@click.option("-t", "--pid-type", multiple=True, required=True)
@with_appcontext
def reindex(pid_type):
    """Reindex all records.

    :param pid_type: Pid type.
    """
    click.secho("Sending records to indexing queue ...", fg="green")

    query = (
        x[0]
        for x in PersistentIdentifier.query.filter_by(
            object_type="rec", status=PIDStatus.REGISTERED
        )
        .filter(PersistentIdentifier.pid_type.in_(pid_type))
        .values(PersistentIdentifier.object_uuid)
    )
    RecordIndexer().bulk_index(query)
    click.secho('Execute "run" command to process the queue!', fg="yellow")


@index.group(chain=True)
@click.option("-q", "--queue", "indexer_ids", multiple=True)
@click.option("-a", "--all-queues", is_flag=True)
def queue(indexer_ids=[], all_queues=False):
    """Manage indexing queue."""


@resultcallback(queue)
@with_appcontext
def process_actions(actions, indexer_ids=[], all_queues=False):
    """Process queue actions."""
    # Resolve indexer IDs
    if all_queues:
        indexers = current_indexer_registry.all().values()
    else:
        indexers = [current_indexer_registry.get(i) for i in indexer_ids]
    # Get queues
    queues = [i.mq_queue for i in indexers]
    # Always add the default indexer queue (for backwards compatibility)
    queues.append(current_app.config["INDEXER_MQ_QUEUE"])
    with establish_connection() as c:
        for queue in queues:
            bound_queue = queue(c)
            for action in actions:
                action(bound_queue)


@queue.command("init")
def init_queue():
    """Initialize indexing queue."""

    def action(queue):
        queue.declare()
        click.secho("Indexing queue has been initialized.", fg="green")
        return queue

    return action


@queue.command("purge")
def purge_queue():
    """Purge indexing queue."""

    def action(queue):
        queue.purge()
        click.secho("Indexing queue has been purged.", fg="green")
        return queue

    return action


@queue.command("delete")
def delete_queue():
    """Delete indexing queue."""

    def action(queue):
        queue.delete()
        click.secho("Indexing queue has been deleted.", fg="green")
        return queue

    return action



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test CLI."""

import uuid

import amqp
import pytest
from celery.messaging import establish_connection
from invenio_db import db
from invenio_pidstore.models import PersistentIdentifier, PIDStatus
from invenio_records import Record
from invenio_search import current_search, current_search_client
from kombu import Queue

from invenio_indexer import cli
from invenio_indexer.api import RecordIndexer
from invenio_indexer.proxies import current_indexer_registry


def test_run(app):
    """Test run."""
    runner = app.test_cli_runner()
    res = runner.invoke(cli.run, [])
    assert 0 == res.exit_code

    runner = app.test_cli_runner()
    res = runner.invoke(cli.run, ["-d", "-c", "2"])
    assert 0 == res.exit_code
    assert "Starting 2 tasks" in res.output


def test_reindex(app):
    """Test reindex."""
    # load records
    with app.test_request_context():
        runner = app.test_cli_runner()

        id1 = uuid.uuid4()
        id2 = uuid.uuid4()
        record1 = Record.create(dict(title="Test 1", recid=1), id_=id1)
        record2 = Record.create(dict(title="Test 2", recid=2), id_=id2)
        PersistentIdentifier.create(
            pid_type="recid",
            pid_value=1,
            object_type="rec",
            object_uuid=id1,
            status=PIDStatus.REGISTERED,
        )
        PersistentIdentifier.create(
            pid_type="recid",
            pid_value=2,
            object_type="rec",
            object_uuid=id2,
            status=PIDStatus.REGISTERED,
        )
        db.session.commit()
        indexer = RecordIndexer()
        index = indexer.record_to_index(record1)

        # Make sure the index doesn't exist at the beginning (it was not
        # preserved by accident from some other tests)
        assert current_search_client.indices.exists(index) is False

        # Initialize queue
        res = runner.invoke(cli.queue, ["init", "purge"])
        assert 0 == res.exit_code

        res = runner.invoke(cli.reindex, ["--yes-i-know", "-t", "recid"])
        assert 0 == res.exit_code
        res = runner.invoke(cli.run, [])
        assert 0 == res.exit_code
        current_search.flush_and_refresh(index)

        # Both records should be indexed
        res = current_search_client.search(index=index)
        assert len(res["hits"]["hits"]) == 2

        # Delete one of the records
        record2 = Record.get_record(id2)
        record2.delete()
        db.session.commit()
        # Destroy the index and reindex
        list(current_search.delete(ignore=[404]))
        res = runner.invoke(cli.reindex, ["--yes-i-know", "-t", "recid"])
        assert 0 == res.exit_code
        res = runner.invoke(cli.run, [])
        assert 0 == res.exit_code
        current_search.flush_and_refresh(index)

        # Check that the deleted record is not indexed
        res = current_search_client.search(index=index)
        assert len(res["hits"]["hits"]) == 1
        assert res["hits"]["hits"][0]["_source"]["title"] == "Test 1"

        # Destroy queue and the index
        res = runner.invoke(cli.queue, ["delete"])
        assert 0 == res.exit_code
        list(current_search.delete(ignore=[404]))


def test_queues_options(app):
    """Test queue sub-command options."""

    cli_runner = app.test_cli_runner()

    users_indexer = RecordIndexer(
        queue=Queue(
            "users",
            exchange=app.config["INDEXER_MQ_EXCHANGE"],
            routing_key="users",
        ),
        routing_key="users",
    )
    files_indexer = RecordIndexer(
        queue=Queue(
            "files",
            exchange=app.config["INDEXER_MQ_EXCHANGE"],
            routing_key="files",
        ),
        routing_key="files",
    )
    current_indexer_registry.register(users_indexer, "users")
    current_indexer_registry.register(files_indexer, "files")
    queues = {
        "default": app.config["INDEXER_MQ_QUEUE"],
        "users": users_indexer.mq_queue,
        "files": files_indexer.mq_queue,
    }

    # Initialize all queues
    res = cli_runner.invoke(cli.queue, ["--all-queues", "init"])
    assert 0 == res.exit_code

    with establish_connection() as c:
        for queue in queues.values():
            assert queue(c).queue_declare(passive=True)

    # Delete all queues
    res = cli_runner.invoke(cli.queue, ["--all-queues", "delete"])
    assert 0 == res.exit_code

    with establish_connection() as c:
        for queue in queues.values():
            with pytest.raises(amqp.exceptions.NotFound):
                queue(c).queue_declare(passive=True)

    # Initialize users queue
    res = cli_runner.invoke(cli.queue, ["--queue", "users", "init"])
    assert 0 == res.exit_code

    with establish_connection() as c:
        assert queues["users"](c).queue_declare(passive=True)

    users_indexer.bulk_index(["123", "456"])
    files_indexer.bulk_index(["file1", "file2", "file3"])

    with establish_connection() as c:
        assert queues["users"](c).queue_declare(passive=True).message_count == 2
        assert queues["files"](c).queue_declare(passive=True).message_count == 3
        assert queues["default"](c).queue_declare(passive=True).message_count == 0

    # Purge files queue
    res = cli_runner.invoke(cli.queue, ["--queue", "files", "purge"])
    assert 0 == res.exit_code

    with establish_connection() as c:
        assert queues["users"](c).queue_declare(passive=True).message_count == 2
        assert queues["files"](c).queue_declare(passive=True).message_count == 0
        assert queues["default"](c).queue_declare(passive=True).message_count == 0

    # Delete files and default queue
    res = cli_runner.invoke(cli.queue, ["--queue", "files", "delete"])
    assert 0 == res.exit_code

    with establish_connection() as c:
        assert queues["users"](c).queue_declare(passive=True).message_count == 2
        with pytest.raises(amqp.exceptions.NotFound):
            queues["files"](c).queue_declare(passive=True)
        with pytest.raises(amqp.exceptions.NotFound):
            queues["default"](c).queue_declare(passive=True)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Record indexer for Invenio."""

from kombu import Exchange, Queue

INDEXER_DEFAULT_INDEX = None
"""Default index to use if no schema is defined."""

INDEXER_MQ_EXCHANGE = Exchange("indexer", type="direct")
"""Default exchange for message queue."""

INDEXER_MQ_QUEUE = Queue("indexer", exchange=INDEXER_MQ_EXCHANGE, routing_key="indexer")
"""Default queue for message queue."""

INDEXER_MQ_ROUTING_KEY = "indexer"
"""Default routing key for message queue."""

INDEXER_MQ_PUBLISH_KWARGS = {}
"""Default message queue producer publishing kwargs.

Passed to ``kombu.Producer:publish``.

.. code-block:: python

    INDEXER_MQ_PUBLISH_KWARGS = {
        "retry": True,
        "retry_policy": {         # Setting for maximum waiting time of ~10min:
            "interval_start": 0,  # First retry immediately,
            "interval_step": 2,   # then increase by 2s for every retry.
            "interval_max": 30,   # but don't exceed 30s between retries.
            "max_retries": 30,    # give up after 30 tries.
        },
    }
"""

INDEXER_REPLACE_REFS = True
"""Whether to replace JSONRefs prior to indexing record."""

INDEXER_BULK_REQUEST_TIMEOUT = 10
"""Request timeout to use in Bulk indexing."""

INDEXER_MAX_BULK_CONSUMERS = 5
"""Maximum number of concurrent consumers for bulk indexing.

This threshold is applied per queue, so each indexing queue would
have a maximum of 5 consumers at the same time.
"""

INDEXER_RECORD_TO_INDEX = "invenio_indexer.utils.default_record_to_index"
"""Provide an implementation of record_to_index function"""

INDEXER_BEFORE_INDEX_HOOKS = []
"""List of automatically connected hooks (function or importable string)."""
